/*****************************************************************************
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: ctrec.c,v 1.9 2000/05/08 20:02:32 kevin Exp $
**  $Log: ctrec.c,v $
**  Revision 1.9  2000/05/08 20:02:32  kevin
**  ANSI C changes
**
**  Revision 1.8  2000/05/04 18:16:34  kevin
**  renamed filter definitions
**
**  Revision 1.7  2000/05/03 08:49:50  kevin
**  Code cleanup
**
**  Revision 1.6  2000/05/02 15:31:47  kevin
**  code cleaning
**
**  Revision 1.5  2000/04/30 11:41:06  kevin
**  Cleaned up debugging code
**
**  Revision 1.4  2000/04/30 10:13:27  kevin
**  Fixed MPI bugs
**
**  Revision 1.3  2000/04/30 04:06:13  kevin
**  Update Raysum i/o routines
**  Fix MPI bug in ctrec (scatter_raysum) that referenced rs_global
**
**  Revision 1.2  2000/04/29 23:24:56  kevin
**  *** empty log message ***
**
**  Revision 1.1.1.1  2000/04/28 13:02:44  kevin
**  Initial CVS import for first public release
**
**
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

/* FILE
 *   ctrec.c		Reconstruct an image from raysums
 *
 * DATE
 *   Aug 84
 *   Jul 99  -- Converted to ANSI C
 *              Added MPI parallel processing
 */

#include "ct.h"

enum {O_INTERP, O_FILTER, O_FILTER_PARAM, O_BACKPROJ, O_VERBOSE, O_TRACE, O_HELP, O_DEBUG, O_VERSION};

static struct option my_options[] =
{
  {"interp", 1, 0, O_INTERP},
  {"filter", 1, 0, O_FILTER},
  {"filter-param", 1, 0, O_FILTER_PARAM},
  {"backproj", 1, 0, O_BACKPROJ},
  {"trace", 1, 0, O_TRACE},
  {"debug", 0, 0, O_DEBUG},
  {"verbose", 0, 0, O_VERBOSE},
  {"help", 0, 0, O_HELP},
  {"version", 0, 0, O_VERSION},
  {0, 0, 0, 0}
};

void 
usage (const char *program)
{
  fprintf(stdout,"usage: %s raysum-file image-file nx-image ny-image [OPTIONS]\n", kbasename(program));
  fprintf(stdout,"Image reconstruction from raysum projections\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"   raysum-file     Input raysum file\n");
  fprintf(stdout,"   image-file      Output image file in SDF2D format\n");
  fprintf(stdout,"   nx-image        Number of columns in output image\n");
  fprintf(stdout,"   ny-image        Number of rows in output image\n");
  fprintf(stdout,"   --interp        Interpolation method during backprojection\n");
  fprintf(stdout,"       nearest     Nearest neighbor interpolation\n");
  fprintf(stdout,"       linear      Linear interpolation\n");
  fprintf(stdout,"       bspline     B-spline interpolation\n");
  fprintf(stdout,"    --filter       Filter name\n");
  fprintf(stdout,"       abs_bandlimit Abs * Bandlimiting (default)\n");
  fprintf(stdout,"       abs_sinc      Abs * Sinc\n");
  fprintf(stdout,"       abs_cos       Abs * Cosine\n");
  fprintf(stdout,"       abs_hamming   Abs * Hamming\n");
  fprintf(stdout,"       shepp         Shepp-Logan\n");
  fprintf(stdout,"       bandlimit     Bandlimiting\n");
  fprintf(stdout,"       sinc          Sinc\n");
  fprintf(stdout,"       cos           Cosine\n");
  fprintf(stdout,"       triangle      Triangle\n");
  fprintf(stdout,"       hamming       Hamming\n");
  fprintf(stdout,"    --backproj     Backprojection Method\n");
  fprintf(stdout,"       trig        Trigometric functions at every point\n");
  fprintf(stdout,"       table       Trigometric functions with precalculated table\n");
  fprintf(stdout,"       diff        Difference method\n");
  fprintf(stdout,"       diff2       Optimized difference method (default)\n");
  fprintf(stdout,"       idiff2      Optimized difference method with integer math\n");
  fprintf(stdout,"    --filter-param Alpha level for Hamming filter\n");
  fprintf(stdout,"    --trace        Set tracing to level\n");
  fprintf(stdout,"         none      No tracing (default)\n");
  fprintf(stdout,"         status    Status tracing\n");
  fprintf(stdout,"         phm       Trace phantom\n");
  fprintf(stdout,"         rays      Trace allrays\n");
  fprintf(stdout,"         plot      Trace plotting\n");
  fprintf(stdout,"         clipping  Trace clipping\n");
  fprintf(stdout,"    --verbose      Turn on verbose mode\n");
  fprintf(stdout,"    --debug        Turn on debug mode\n");
  fprintf(stdout,"    --version      Print version\n");
  fprintf(stdout,"    --help         Print this help message\n");
  exit(1);  
}


#ifdef MPI_CT
static void mpi_scatter_rs (RAYSUM *rs_global, RAYSUM *rs_local, const int debug);
#endif
static void print_raysum_info(const RAYSUM *rs);

int 
main (const int argc, char *const argv[])
{
  IMAGE *im_global = NULL;
  RAYSUM *rs_global = NULL;
  char *rs_name, *im_filename = NULL;
  char remark[MAXREMARK];
  char filt_name[80];
  int nx, ny;
  double time_start = 0, time_end = 0;
  char *endptr;
  int opt_verbose = 0;
  int opt_debug = 0;
  int opt_trace = TRACE_NONE;
  double opt_filter_param = 1;
  int opt_filter = FILTER_ABS_BANDLIMIT;
  int opt_interp = I_LINEAR;
  int opt_interp_param = 1;
  BackprojType opt_backproj = O_BPROJ_DIFF2;
#ifdef MPI_CT
  IMAGE *im_local;
  RAYSUM *rs_local;
  int mpi_argc = argc;
  char **mpi_argv = (char **) argv;
  int mpi_nview, mpi_ndet;
  double mpi_detinc, mpi_rotinc, mpi_phmlen;
  double mpi_t1, mpi_t2, mpi_t, mpi_t_g;
  int ix;

  MPI_Init(&mpi_argc, &mpi_argv);
  MPI_Comm_dup (MPI_COMM_WORLD, &mpi_ct.comm);
  MPI_Comm_size(mpi_ct.comm, &mpi_ct.nproc);
  MPI_Comm_rank(mpi_ct.comm, &mpi_ct.my_rank);

  if (mpi_ct.nproc > MPI_MAX_PROCESS) {
    sys_error(ERR_FATAL, "Number of mpi processes (%d) exceeds max processes (%d)",
	      mpi_ct.nproc, MPI_MAX_PROCESS);
    exit(1);
  }
#endif

#ifdef MPI_CT
  time_start = MPI_Wtime();
#else
  time_start = td_current_sec();
#endif

#ifdef MPI_CT
  if (mpi_ct.my_rank == 0) {
#endif
    while (1) {
      int c = getopt_long(argc, argv, "", my_options, NULL);
      char *endptr = NULL;
      
      if (c == -1)
	break;
      
      switch (c)
	{
	case O_INTERP:
	  opt_interp = opt_set_interpolation(optarg, argv[0]);
	  break;
	case O_FILTER:
	  opt_filter = opt_set_filter(optarg, argv[0]);
	  break;
	case O_BACKPROJ:
	  opt_backproj = opt_set_backproj(optarg, argv[0]);
	  break;
	case O_FILTER_PARAM:
	  opt_filter_param = strtod(optarg, &endptr);
	  if (endptr != optarg + strlen(optarg)) {
	    usage(argv[0]);
	    exit(1);
	  }
	  break;
	case O_VERBOSE:
	  opt_verbose = 1;
	  break;
	case O_DEBUG:
	  opt_debug = 1;
	  break;
	case O_TRACE:
	  opt_trace = opt_set_trace(optarg, argv[0]);
	  break;
        case O_VERSION:
#ifdef VERSION
	  fprintf(stdout, "Version %s\n", VERSION);
#else
	  fprintf(stderr, "Unknown version number");
#endif
	  exit(0);
	case O_HELP:
	case '?':
	  usage(argv[0]);
	  exit(0);
	default:
	  usage(argv[0]);
	  exit(1);
	}
    }
  
    if (optind + 4 != argc) {
      usage(argv[0]);
      exit(1);
    }

    rs_name = argv[optind];
  
    im_filename = argv[optind + 1];
  
    nx = strtol(argv[optind + 2], &endptr, 10);
    ny = strtol(argv[optind + 3], &endptr, 10);
  
    if (opt_filter == FILTER_G_HAMMING || opt_filter == FILTER_ABS_G_HAMMING)
      sprintf (filt_name, "%s: alpha = %.2f",
	       name_of_filter (opt_filter), opt_filter_param); 
    else
      sprintf (filt_name, "%s", name_of_filter (opt_filter));
  
    sprintf (remark, "Reconstruct: %dx%d, %s, %s, %s",
	     nx, ny, filt_name, name_of_interpolation (opt_interp), name_of_backproj(opt_backproj));
  
    if (opt_verbose)
      fprintf (stdout, "%s\n", remark);
#ifdef MPI_CT
  }
#endif

#ifdef MPI_CT
  if (mpi_ct.my_rank == 0) {
    rs_global = raysum_open (rs_name);
    raysum_read (rs_global);
    if (opt_verbose)
      print_raysum_info(rs_global);

    mpi_ndet = rs_global->ndet;
    mpi_nview = rs_global->nview;
    mpi_detinc = rs_global->det_inc;
    mpi_phmlen = rs_global->phmlen;
    mpi_rotinc = rs_global->rot_inc;
  }

  mpi_t1 = MPI_Wtime();
  MPI_Bcast(&opt_verbose, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_debug, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_trace, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_filter, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_interp, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_filter_param, 1, MPI_DOUBLE, 0, mpi_ct.comm);
  MPI_Bcast(&opt_interp_param, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&opt_backproj, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&mpi_ndet, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&mpi_nview, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&mpi_detinc, 1, MPI_DOUBLE, 0, mpi_ct.comm);
  MPI_Bcast(&mpi_phmlen, 1, MPI_DOUBLE, 0, mpi_ct.comm);
  MPI_Bcast(&mpi_rotinc, 1, MPI_DOUBLE, 0, mpi_ct.comm);
  MPI_Bcast(&nx, 1, MPI_INT, 0, mpi_ct.comm);
  MPI_Bcast(&ny, 1, MPI_INT, 0, mpi_ct.comm);
  if (opt_verbose) {
    mpi_t2 = MPI_Wtime();
    mpi_t = mpi_t2 - mpi_t1;
    MPI_Reduce(&mpi_t, &mpi_t_g, 1, MPI_DOUBLE, MPI_MAX, 0, mpi_ct.comm);
    if (mpi_ct.my_rank == 0)
      printf("Time to Bcast vars = %f secs, Max time = %f\n", mpi_t, mpi_t_g);
  }

  mpi_ct_calc_work_units(mpi_nview);

  rs_local = raysum_create (NULL, mpi_ct.local_work_units[mpi_ct.my_rank], mpi_ndet);

  rs_local->ndet = mpi_ndet;
  rs_local->nview = mpi_nview;
  rs_local->det_inc = mpi_detinc;
  rs_local->phmlen = mpi_phmlen;
  rs_local->rot_inc = mpi_rotinc;

  if (opt_verbose)
    mpi_t1 = MPI_Wtime();
  mpi_scatter_rs(rs_global, rs_local, opt_debug);
  if (opt_verbose) {
    mpi_t2 = MPI_Wtime();
    mpi_t = mpi_t2 - mpi_t1;
    MPI_Reduce(&mpi_t, &mpi_t_g, 1, MPI_DOUBLE, MPI_MAX, 0, mpi_ct.comm);
    if (mpi_ct.my_rank == 0)
      printf("Time to scatter rs = %f secs, Max time = %f sec\n", mpi_t, mpi_t_g);
  }

  if (mpi_ct.my_rank == 0) {
    im_global = image_create (im_filename, nx, ny);
    sdf_add_label (LT_HISTORY, rs_global->remark, rs_global->calctime, im_global->dfp_2d->dfp);
    sdf_add_empty_label (im_global->dfp_2d->dfp);
  }
  im_local = image_create (NULL, nx, ny);

#else
  rs_global = raysum_open (rs_name);
  raysum_read (rs_global);
  if (opt_verbose)
    print_raysum_info(rs_global);

  im_global = image_create (im_filename, nx, ny);
  sdf_add_label (LT_HISTORY, rs_global->remark, rs_global->calctime, im_global->dfp_2d->dfp);
  sdf_add_empty_label (im_global->dfp_2d->dfp);
#endif

#ifdef MPI_CT
  mpi_t1 = MPI_Wtime();
  image_reconst (im_local, rs_local, opt_filter, opt_filter_param, 
		 opt_interp, opt_interp_param, opt_backproj, opt_trace);
  mpi_t2 = MPI_Wtime();
  mpi_t = mpi_t2 - mpi_t1;
  MPI_Reduce(&mpi_t, &mpi_t_g, 1, MPI_DOUBLE, MPI_MAX, 0, mpi_ct.comm);
  if (mpi_ct.my_rank == 0 && opt_verbose)
    printf("Time to reconstruct = %f, Max time = %f\n", mpi_t, mpi_t_g);
#else
  image_reconst (im_global, rs_global, opt_filter, opt_filter_param, 
		 opt_interp, opt_interp_param, opt_backproj, opt_trace);
#endif

#ifdef MPI_CT
  if (mpi_ct.my_rank == 0) {
    raysum_close (rs_global);
  }

  if (opt_verbose)
    mpi_t1 = MPI_Wtime();

  for (ix = 0; ix < im_local->nx; ix++) {
    void *recvbuf = NULL;
    if (mpi_ct.my_rank == 0)
      recvbuf = im_global->v[ix];

    MPI_Reduce(im_local->v[ix], recvbuf, im_local->ny, MPI_FLOAT, MPI_SUM, 0, mpi_ct.comm);
  }

  if (opt_verbose) {
    mpi_t2 = MPI_Wtime();
    mpi_t = mpi_t2 - mpi_t1;
    MPI_Reduce(&mpi_t, &mpi_t_g, 1, MPI_DOUBLE, MPI_MAX, 0, mpi_ct.comm);
    if (mpi_ct.my_rank == 0)
      printf("Time to reduce image = %f secs, max time = %f\n", mpi_t, mpi_t_g);
  }
  if (mpi_ct.my_rank == 0) {
    strncpy (im_global->remark, remark, sizeof(im_global->remark));
    time_end = MPI_Wtime();
    im_global->calctime = time_end - time_start;
    image_save (im_global);
    if (opt_verbose)
      fprintf (stdout, "Time active = %.2f\n", im_global->calctime);
    }
#else  
  raysum_close (rs_global);
  strncpy (im_global->remark, remark, sizeof(im_global->remark));
  time_end = td_current_sec();
  im_global->calctime = time_end - time_start;
  image_save (im_global);
  if (opt_verbose)
    fprintf (stdout, "Time active = %.2f secs\n", im_global->calctime);
#endif

#ifdef MPI_CT
	MPI_Finalize();
#endif

	return (0);
}


#ifdef MPI_CT
static void mpi_scatter_rs (RAYSUM *rs_global, RAYSUM *rs_local, const int opt_debug)
{
  int iproc;
  int end_work_unit;
  int iw = 0;

  if (mpi_ct.my_rank == 0) {
    for (iproc = 0; iproc < mpi_ct.nproc; iproc++) {
      end_work_unit = mpi_ct.start_work_unit[iproc] + mpi_ct.local_work_units[iproc]  - 1;

      for (iw = mpi_ct.start_work_unit[iproc]; iw <= end_work_unit; iw++) {
	MPI_Send(&rs_global->view[iw]->ndet, 1, MPI_INT, iproc, 0, mpi_ct.comm);
	MPI_Send(&rs_global->view[iw]->view_angle, 1, MPI_DOUBLE, iproc, 0, mpi_ct.comm);
	MPI_Send(rs_global->view[iw]->detval, rs_global->ndet, MPI_FLOAT, iproc, 0, mpi_ct.comm);
      }
    }
  }

  if (opt_debug)
    fprintf(stdout, "Receiving rs data in process %d\n", mpi_ct.my_rank);

  end_work_unit = mpi_ct.local_work_units[mpi_ct.my_rank] - 1;
  for (iw = 0; iw <= end_work_unit; iw++) {
    MPI_Status status;

    MPI_Recv(&rs_local->view[iw]->ndet, 1, MPI_INT, 0, 0, mpi_ct.comm, &status);
    MPI_Recv(&rs_local->view[iw]->view_angle, 1, MPI_DOUBLE, 0, 0, mpi_ct.comm, &status);
    MPI_Recv(rs_local->view[iw]->detval, rs_local->ndet, MPI_FLOAT, 0, 0, mpi_ct.comm, &status);
  }
  rs_local->nview = mpi_ct.local_work_units[mpi_ct.my_rank];
  if (opt_debug) {
    MPI_Barrier(MPI_COMM_WORLD);
    fprintf(stdout, "Done with mpi_scatter_rs in process %2d\n", mpi_ct.my_rank);
  }
}

#endif

static void print_raysum_info(const RAYSUM *rs)
{
  printf ("Number of detectors: %d\n", rs->ndet);
  printf ("    Number of views: %d\n", rs->nview);
  printf ("             Remark: %s\n", rs->remark);
  printf ("             phmlen: %f\n", rs->phmlen);
  printf ("          det_start: %f\n", rs->det_start);
  printf ("            det_inc: %f\n", rs->det_inc);
  printf ("          rot_start: %f\n", rs->rot_start);
  printf ("            rot_inc: %f\n", rs->rot_inc);
}
